import json
import string
from enum import Enum
from typing import Any, Dict

import websocket

from FayePython.channel import Channel


class ClientStatus(Enum):
    UNCONNECTED = "unconnected"
    HANDSHAKING = "handshaking"
    CONNECTING = "connecting"
    CONNECTED = "connected"
    DISCONNECTED = "disconnected"


class Client:
    BAYEUX_VERSION = "1.0"

    def __init__(self, url: str) -> None:
        print(f"New client created for {url}")  # info

        self._url = url

        self._client_id = None
        self._message_id = 0
        self._state = ClientStatus.UNCONNECTED

        self._response_callbacks = {}
        self.subscribes = {}

        self.wsapp = websocket.WebSocketApp(
            url, on_open=self.connect, on_message=self._receive_message
        )

    def handshake(self) -> None:
        if self._state != ClientStatus.UNCONNECTED:
            return

        self._state = ClientStatus.HANDSHAKING

        print(f"Initiating handshake with {self._url}")  # info

        self._send_message(
            {
                "channel": Channel.META_HANDSHAKE.value,
                "version": self.BAYEUX_VERSION,
                "supportedConnectionTypes": ["websocket"],
            },
            self._on_handshake,
        )

    def connect(self, wsapp) -> None:
        if self._state == ClientStatus.UNCONNECTED:
            return self.handshake()

        self._state = ClientStatus.CONNECTING

        print(f"Initiating connection for {self._client_id}")

        self._send_message(
            {
                "channel": Channel.META_CONNECT.value,
                "clientId": self._client_id,
                "connectionType": "websocket",
            },
            self._cycle_connection,
        )

    def disconnect(self) -> None:
        if self._state != ClientStatus.CONNECTED:
            return

        self._state = ClientStatus.DISCONNECTED

        print(f"Disconnecting for {self._client_id}")  # info

        self._send_message(
            {"channel": Channel.META_DISCONNECT.value, "clientId": self._client_id},
            self._on_disconnect,
        )

    def subscribe(self, subscription: str, callback) -> None:
        # self.subscribes[subscription] = callback

        print(
            f"Client {self._client_id} attempting to subscribe to {subscription}"
        )  # info

        self._send_message(
            {
                "channel": Channel.META_SUBSCRIBE.value,
                "clientId": self._client_id,
                "subscription": subscription,
            }
        )

    # def unsubscribe(self, subscription: str) -> None:
    #     del self.subscribes[subscription]

    #     print(f"Client {self._client_id} attempting to unsubscribe from {subscription}")  # info

    #     self._send_message({
    #         'channel': Channel.META_UNSUBSCRIBE.value,
    #         'clientId': self._client_id,
    #         'subscription': subscription
    #     })

    def _send_message(self, message, callback=None) -> None:
        message["id"] = self._generate_message_id()

        if callback != None and message["channel"] != Channel.META_SUBSCRIBE.value:
            self._response_callbacks[message["id"]] = callback
        self.wsapp.send(json.dumps(message))

    def _receive_message(self, wsapp, message):
        data = json.loads(message)
        for i in data:
            if i["id"] in self._response_callbacks.keys():
                self._response_callbacks[i["id"]](i)
                del self._response_callbacks[i["id"]]
            if i["channel"] in self.subscribes.keys():
                self.subscribes[i["channel"]](i)

    def _base36encode(
        self, number: int, alphabet: str = string.digits + string.ascii_lowercase
    ) -> str:
        if number < 0:
            return "-" + self._base36encode(-number)

        value = ""

        while number != 0:
            number, index = divmod(number, len(alphabet))
            value = alphabet[index] + value

        return value or "0"

    def _generate_message_id(self) -> str:
        self._message_id += 1
        if self._message_id >= 2**32:
            self._message_id = 0
        return self._base36encode(self._message_id)

    def _on_handshake(self, response: Dict[str, Any]) -> None:
        if response["successful"]:
            self._client_id = response["clientId"]

            print(f"Handshake successful: {self._client_id}")  # info

            for key, value in self.subscribes.items():
                self.subscribe(key, value)

            self.connect(None)
        else:
            print("Handshake unsuccessful")  # info

            self._state = ClientStatus.UNCONNECTED

    def _cycle_connection(self, response: Dict[str, Any]) -> None:
        if response["successful"]:
            self._state = ClientStatus.CONNECTED

            print(f"Reconnecting for {self._client_id}")  # info

            self.connect(None)
        else:
            self.disconnect()

    def _on_disconnect(self, response: Dict[str, Any]) -> None:
        self.wsapp.close()
